#ifndef CAMCLAYCOUP_H
#define CAMCLAYCOUP_H

#include "iotools.h"
#include "alias.h"
#include "strretalg.h"
struct matrix;
struct vector;
struct atsel;

/**
   This class defines modified cam-clay material model
   for coupling with moisture transfer. The material is defined by 
   three usual camclay material constants :
   m - is frictional constant and it depends on the frictional angle phi
   lambda - slope of the normal consolidationa line
   kappa  - slope of the swelling line

   Additionally, there are three parameters which controls
   evolution of hardening with respect to degree of saturation and suction
   c1_tilde, c2 - controls the ration of specific volumes v/v_sat (partially sturated/fully saturated)
   ks - controls apparent adhesion (intersection of yield surface with p axis)
   
   In addition to these material constants, several initial values 
   have to be specified at each integration point in the following order :
   v_kappa1   - initial specific volume at the reference pressure p1 after 
                unloading from initial consolidation pressure p_c0 (v_kappa0)
   p1         - reference pressure (p_1), it must be negative
   p_c0       - initial consolidation pressure, it must be negative
   iepsp_1    -
      .        \
      .         \ components of initial plastic strains
      .         /
   iepsp_ncomp -  
   
   Order of the eqother array components
   id                | description
   ------------------+------------------------------------------------
   <0;  ncompstr-1>  | plastic strains 
    ncompstr         | gamma - consistency parameter 
    ncompstr+1       | hardening parameter p_c - consolidation pressure
    ncompstr+2       | hardening parameter p_s - apparent adhesion
    ncompstr+3       | v_lambda1 - initial value of specific volume on the NCL at reference pressure p1
        .            | v_ini - initial specific volume
        .            | i1s - mean stress  
        .            | j2s - the second invariant of stress deviator
        .            | epsv - total volume strain 
    ncompstr+7       | epsvp - plastic volume strain  
   

   10.2012
   TKo
*/
class camclaycoup
{
 public:
  camclaycoup (void);
  ~camclaycoup (void);
  void read (XFILE *in);
  void initval(long ipp, long ido);
  double cohesion(vector &qtr);
  double yieldfunction (vector &sig, vector &q);
  void deryieldfsigma (vector &sig, vector &q, vector &dfds);
  void dderyieldfsigma (matrix &ddfds);
  void derpotsigma (vector &sig, vector &q, vector &dgds);
  void deryieldfq(vector &sig, vector &q, vector &dq);
  void deryieldfdqdq(matrix &dfq);
  void deryieldfdsdq(matrix &dfdsdqt);
  void der_q_gamma(long ipp, long ido, vector &sig, vector &qtr, vector &dqdg);
  double plasmodscalar(long ipp, long ido, vector &sig, vector &qtr);
  void updateq(long ipp, long ido, vector &eps, vector &epsp, vector &sig, vector &q);
//  void updateq(vector &epsp, strastrestate ssst, vector &q);
  void matstiff (matrix &d, long ipp,long ido);
  void nlstresses (long ipp,long im,long ido);
  void updateval (long ipp,long ido);
  void giveirrstrains (long ipp, long ido, vector &epsp);
  double give_consparam (long ipp, long ido);
  double give_preconspress(long ipp, long ido);
  double give_saturation_degree(long ipp, long ido);
  double give_virgporosity(long ipp, long ido);
  double give_iniporosity(long ipp, long ido);
  void give_reqnmq(long *anmq);
  void changeparam (atsel &atm, vector &val);

  /// frictional constant
  double m;

  /// slope of normal consolidation line
  double lambda;

  /// slope of swelling line
  double kappa;

  /// first parameter for control of the v/v_sat ratio evolution
  double c1_tilde;

  /// second parameter for control of the v/v_sat ratio evolution
  double c2;

  /// coefficent of apparent adhesion due to suction pressure
  double ks;

  ///  stress return algorithm
  strretalg sra;
};

#endif
