#ifndef HYPOPLAST_UNSAT_EXP_THERM_H
#define HYPOPLAST_UNSAT_EXP_THERM_H

#include "galias.h"
#include "gfunct.h"
#include "generalmod.h"
#include "strretalg.h"

struct XFILE;
struct matrix;
struct vector;
struct atsel;


/**
  The class defines hypoplastic model for unsaturated expansive soils
  with double porosity structure and thermal effects according to D. Masin.

  Ordering of eqother/other array:
  --------------------------------

  History of qunatities handeled by MEFEL:
  - previous strain vector               /eqother[0] - eqother[ncompstr-1]/,
  - previous stress vector               /eqother[ncompstr] - eqother[2*ncompstr-1]/,
  - initial increments of 
    normal stress components 
    due to pore pressure change          /eqother[2*ncompstr - eqother[2*ncompstr+2],
    due to pore pressure change          /eqother[2*ncompstr - eqother[2*ncompstr+3],
    due to pore pressure change          /eqother[2*ncompstr - eqother[2*ncompstr+4],
    the number of model evaluations      /eqother[2*ncompstr - eqother[2*ncompstr+5],

  Internal variables required and handeled by hypoplasticity model:
  - actual void ratio       e            /eqother[2*ncompstr+3]/,
  - previous/actual suction s            /eqother[2*ncompstr+3+1]/,
  - degree of saturation    S_r          /eqother[2*ncompstr+3+2]/,
  - temperatutre            T            /eqother[2*ncompstr+3+3]/,
  -                         em           /eqother[2*ncompstr+3+4]/,
  -                         eM           /eqother[2*ncompstr+3+5]/,
  -                         SrM          /eqother[2*ncompstr+3+6]/,
  -                         ascan        /eqother[2*ncompstr+3+7]/,
  -                         re           /eqother[2*ncompstr+3+8]/,
*/
class hypoplast_unsat_exp_therm
{
 public:
  hypoplast_unsat_exp_therm();
  ~hypoplast_unsat_exp_therm();
  void read(XFILE *in);
  void print(FILE *out);
  void initval (long ipp,long ido);
  void matstiff (matrix &d,long ipp,long ido);
  void nlstressesincr (long ipp, long im, long ido);
  void givestressincr (long lcid, long ipp, long im, long ido, long fi, vector &sig);
  void nlstresses (long ipp, long im, long ido);
  long rkf23(vector &eps, vector &sig, vector &statev, vector &deps, double &dtsub, long &neval);
  long rkf23bs(vector &eps, vector &sig, vector &statev, vector &deps, double &dtsub, long &neval);
  long rkf34(vector &eps, vector &sig, vector &statev, vector &deps, double &dtsub, long &neval);
  long rkf45(vector &eps, vector &sig, vector &statev, vector &deps, double &dtsub, long &neval);
  double dstep_red(long ipp, long im, long ido);
  void updateval (long ipp,long im,long ido);
  void give_reqnmq(long *anmq);
  double give_virgporosity(long ipp, long ido);
  double give_iniporosity(long ipp, long ido);
  void changeparam (atsel &atm,vector &val);
  
  double phideg;      // critical state friction angle
  double lam_star;    // slope of the normal compression lines (isotropic/oedometric/critical state line) of a saturated soil 
                      // in the ln(p^M/p_r) vs ln(1+e) plane
  double kap_star;    // slope of the macrostructural isotropic unloading line of a saturated soil in the ln(p^M/p_r) vs ln(1+e) plane (saturated state)
  double n_star;      // position of the isotropic normal compression line, i.e. the value of ln(1+e) for p^M = p_r = 1kPa (saturated state)
  double nu;          // Poisson's ratio ???
  double n;           // defines position of the normal compression line with respect to suction  N(s) = n_star + n*<ln(s/s_e)>
  double l;           // defines slope of the normal compression line with respect to suction  lambda(s) = lam_star + l*<ln(s/s_e)>
  double nt;          // defines position of the normal compression line with respect to temperature ???
  double lt;          // defines slope of the normal compression line with respect to temperature ???
  double m;           // controls the dependency of wetting-induced collapse on the overconsolidation ratio and the macropore occlusion 
                      // by microporosity on relative void ratio
  double alpha_s;     // thermal dilatancy coefficient ???
  double kap_m;       // dependency of microstructural swelling/shrinkage on the microstructural effective stress (saturated effective stress)
  double sm_star;     // suction corresponding to em_star
  double em_star;     // void ratio of micropores for fully saturated micropores (S^m_r) and dry macropores (S^M_r)
  double csh;         // 
  double s_airentry0; // the air entry value of suction for the reference macrostructural void ratio e_0^M (fully saturated microstructure)
  double em0;         // reference macrostructural void ratio e_0
  double tr;
  double at;
  double bt;
  double aer;
  double p_t;  

  answertype pr_suc_fl;
  gfunct suc_fn;
  answertype pr_tempr_fl;
  gfunct tempr_fn;

  long nparams; //number of parameters
  long nstatev;  //number of state variables
  long ngstrain;  //number of generalised strain tensor components
  long ngstress;  //number of generalised stress tensor components
  double *params;
  int nparstatev[4];
  strretalg sra;

  Hypoplasti_unsat_expansive_thermal huet;
};

#endif
